local M = {}

local newDebug = require("debugutil")
local mathutil = require("mathutil")

local pid = {}
pid.__index = pid;

function newPid(config)
    config = config or {}
    
    local obj = {
        kPIn = config.kPIn or config.kP or 0,
        kPOut = config.kPOut or config.kP or 0,
        kIIn = config.kIIn or config.kI or 0,
        kIOut = config.kIOut or config.kI or 0,
        kD = config.kD or 0,
        minOutput = config.minOutput or -math.huge,
        maxOutput = config.maxOutput or math.huge,
        scale = config.scale or 1,
        name = config.name or "pid",
        dThreshold = config.integralLockThreshold,
        dThresholdReset = config.integralLockAutoReset == true,
        debug = config.debug == true,
        output = 0,
        
        lastValue = nil,
        lastError = nil,
        integral = 0,
        lockIntegral = false,
        dynamicIntegralLock = false
    }
    
    obj.minIntegral = config.minIntegral or obj.minOutput
    obj.maxIntegral = config.maxIntegral or obj.maxOutput
    obj.debugutil = newDebug("pid_" .. obj.name, obj.debug)
    
    setmetatable(obj, pid)
    return obj
end

function pid:get(target, actual, dt)
    local err = (target - actual) * self.scale
    local delta = self.lastError and (err - self.lastError) * self.scale or 0
    local pFactor = (err * actual > 0 and self.kPOut or self.kPIn) * err
    local dFactor = self.kD * delta / dt
    local lockIntegral = self.lockIntegral or self.dynamicIntegralLock
    
    if self.dThreshold and math.abs(delta) > self.dThreshold then
        self.dynamicIntegralLock = true
        
        if self.dThresholdReset then
            self.integral = 0
        end
    else
        self.dynamicIntegralLock = false
    end
    
    if not lockIntegral then
        self.integral = mathutil.clamp(self.integral + err * (err * actual > 0 and self.kIOut or self.kIIn) * dt, self.minIntegral, self.maxIntegral)
    end
    
    local iFactor = self.integral
    
    self.output = mathutil.clamp(pFactor + iFactor - dFactor, self.minOutput, self.maxOutput)
    self.lastValue = actual
    self.lastError = err
    
    if self.debug then
        self.debugutil:putValue("Target", target, 3)
        self.debugutil:putValue("Actual", actual, 3)
        self.debugutil:putValue("Delta", delta / self.scale, 3)
        self.debugutil:putValue("Error", err / self.scale, 3)
        self.debugutil:putValue("Integral", self.integral, 3)
        self.debugutil:putValue("Last Value", self.lastValue, 3)
        self.debugutil:putValue("Last Error", self.lastError, 3)
        self.debugutil:putValue("P Factor", pFactor, 3)
        self.debugutil:putValue("I Factor", iFactor, 3)
        self.debugutil:putValue("D Factor", -dFactor, 3)
        self.debugutil:putValue("I Lock", lockIntegral and "True" or "False")
        self.debugutil:putValue("Output", self.output, 3)
    end
    
    return self.output
end

function pid:peek()
    return self.output
end

function pid:setConfig(config)
    config = config or {}
    
    for k, v in pairs(config) do
        self[k] = v
    end
    
    self.debugutil.enabled = self.debug == true
end

function pid:reset()
    self.integral = 0
    
    return self
end

function pid:setLockIntegral(lockIntegral)
    self.lockIntegral = lockIntegral or false
    
    return self
end

M.newPid = newPid

return M