local M = {}
M.outputPorts = {[1] = true}
M.deviceCategories = {engine = true}
M.requiredExternalInertiaOutputs = {1}

-- imports
local mathutil = require("mathutil")
local debugutil = require("debugutil")

-- function aliases
local abs = math.abs
local min = math.min
local max = math.max
local clamp = mathutil.clamp
local sign = mathutil.sign
local maprange = mathutil.maprange

-- constants
local constants = {
    avToRPM = 9.549296596425384,
    dynamicFrictionCoef = 0.1047197177,
}

-- prototype
local motor = {}
motor.__index = motor

function motor:velocityUpdate() end -- just a nop

function motor:torqueUpdate(dt)
    self.motorAV = self.outputAV1
    
    local avSign = sign(self.motorAV)
    local outputRPM = self.motorAV * constants.avToRPM
    local motorDir = self.outputTorqueDirection
    local friction = self.friction
    local dynamicFriction = self.dynamicFriction
    local maxMotorTorque = self:getMaxMotorTorque(self.outputTorqueDemand < 0)
    local maxInstantTorque = maxMotorTorque - friction - (dynamicFriction * outputRPM * constants.dynamicFrictionCoef)
    local actualTorque
    
    if self.outputTorqueDemand < 0 then
        actualTorque = -avSign * min(abs(self.outputTorqueDemand), self:getMaxRegenTorque())
    else
        actualTorque = self.outputTorqueDemand * motorDir
    end
    
    actualTorque = clamp(actualTorque, -maxInstantTorque, maxInstantTorque)
    
    local frictionTorque = abs(friction * avSign + dynamicFriction * self.motorAV)
    
    frictionTorque = min(frictionTorque, abs(self.motorAV) * self.motorInertia * 2000) * avSign
    
    self.motorTorque = actualTorque
    self.outputTorque1 = actualTorque - frictionTorque
    
    local motorPowerCoef = self.outputTorqueDemand < 0 and 1 or self:getMotorPowerCoef(self.motorAV)
    local effectiveMotorAV = max(self.motorMinNoStallAV, abs(self.motorAV))
    
    self.motorPower = motorPowerCoef * sign(self.outputTorqueDemand) * abs(mathutil.torqueToKw(self.outputTorque1, effectiveMotorAV))
    
    self.debug:putValue("Motor RPM (AV)", self.debug:formatNumber(self.motorAV * constants.avToRPM, 2) .. " (" .. self.debug:formatNumber(self.motorAV, 2) .. ")")
    self.debug:putValue("Motor Torque (Nm)", self.motorTorque, 3)
    self.debug:putValue("Motor Power (kW)", self.motorPower, 3)
    self.debug:putValue("Output Torque (Nm)", self.outputTorque1, 3)
end

function motor:calculateInertia()
    local outputInertia = 0
    local cumulativeGearRatio = 1
    local maxCumulativeGearRatio = 1
    
    if self.children and #self.children > 0 then
        outputInertia = self.children[1].cumulativeInertia
        cumulativeGearRatio = self.children[1].cumulativeGearRatio
        maxCumulativeGearRatio = self.children[1].maxCumulativeGearRatio
    end

    self.cumulativeInertia = outputInertia
    self.cumulativeGearRatio = cumulativeGearRatio
    self.maxCumulativeGearRatio = maxCumulativeGearRatio
end

function motor:getMaxMotorTorque(overrideOrIgnore)
    local motorAV = max(abs(self.motorAV), 1e-6)
    local maxPower = self.motorPowerRating
    
    if type(overrideOrIgnore) == "number" then
        maxPower = min(overrideOrIgnore, maxPower)
    elseif (type(overrideOrIgnore) ~= "boolean" or overrideOrIgnore == false) and type(self.maxMotorPowerOverride) == "number" then
        maxPower = self.maxMotorPowerOverride
    end
    
    local maxTorque = mathutil.kwToTorque(maxPower, motorAV)
    
    return min(maxTorque, self.motorTorqueRating)
end

function motor:getMaxRegenTorque(maxPower)
    local motorAV = self.motorAV
    local maxTorque = self:getMaxMotorTorque(maxPower or true)
    
    return maprange(motorAV, 0, self.motorMinFullRegenAV, 0, maxTorque)
end

function motor:getMotorPowerCoef(motorAV)
    return maprange(abs(motorAV), 0, self.motorMinNoStallAV, self.stallPowerCoef, 1)
end

function motor:setMotorPowerOverride(maxPowerOverride)
    self.maxMotorPowerOverride = maxPowerOverride
end

function motor:setRequestedOutputTorque(requestedTorque)
    self.outputTorqueDemand = requestedTorque
end

function motor:setMotorDirection(direction)
    self.outputTorqueDirection = (direction == -1) and -1 or 1
end

function motor:validate()
    if not self.children or #self.children < 1 then
      self.clutchChild = {torqueDiff = 0}
    elseif #self.children == 1 and self.children[1].deviceCategories.clutchlike then
      self.clutchChild = self.children[1]
      self.invInertia = 1 / (self.motorInertia + (self.clutchChild.additionalEngineInertia or 0))
      self.halfInvInertia = self.invInertia * 0.5
    end
  
    return true
  end

function motor:reset(jbeamData)
    self.outputAV1 = 0
    self.motorAV = 0
    self.inputAV = 0
    self.outputTorque1 = 0
    self.outputTorqueDemand = 0
    self.outputTorqueDirection = 1
    self.motorPower = 0
    self.motorTorque = 0
    self.motorInertia = jbeamData.motorInertia or 0.1
end

local function new(jbeamData)
    local device = {
        deviceCategories = shallowcopy(M.deviceCategories),
        requiredExternalInertiaOutputs = shallowcopy(M.requiredExternalInertiaOutputs),
        outputPorts = shallowcopy(M.outputPorts),
        name = jbeamData.name,
        type = jbeamData.type,
        
        friction = jbeamData.friction or 2,
        dynamicFriction = jbeamData.dynamicFriction or 0,
        cumulativeInertia = 1,
        cumulativeGearRatio = 1,
        maxCumulativeGearRatio = 1,
        outputAV1 = 0,
        motorAV = 0,
        inputAV = 0,
        outputTorque1 = 0,
        outputTorqueDemand = 0,
        outputTorqueDirection = 1,
        motorPower = 0,
        motorTorque = 0,
        
        motorTorqueRating = jbeamData.torqueRating or 150, -- Nm
        motorPowerRating = jbeamData.powerRating or 50,    -- kW
        maxMotorPowerOverride = nil,
        motorMaxAV = jbeamData.maxAV or 1780,  -- rads/s
        motorMinAV = jbeamData.minAV or -(jbeamData.maxAV or 1780), -- rad/s
        motorMinFullRegenAV = jbeamData.minFullRegenAV or 250,
        motorMinNoStallAV = jbeamData.minNoStallAV or 15,
        motorStallPowerCoef = jbeamData.stallPowerCoef or 1.5,
        motorInertia = jbeamData.motorInertia or 0.1,
        debug = debugutil(jbeamData.name or "electricMotor", jbeamData.enableDebug),
    }
    setmetatable(device, motor)
    
    device.invInertia = 1 / device.motorInertia
    device.halfInvInertia = device.invInertia * 0.5
    
    return device
end

M.new = new

return M