local M = {}
local prespooler = {}
prespooler.__index = prespooler

-- Imports
local debugutil = require("debugutil")("Turbo Regen System", true)
local mathutil = require("mathutil")

-- Constants
local constants = {
    -- Conversion rates
    avToRPM = 9.549296596425384,
    rpmToAV = 1.0 / 9.549296596425384,
}

-- Aliases
local abs = math.abs
local min = math.min
local max = math.max
local maprange = mathutil.maprange
local sign = mathutil.sign
local fnum = function(v, p) return debugutil:formatNumber(v, p) end

function prespooler:updateGFX(dt)
    local turboRPM = electrics.values[self.turboElectricsRPMName] or 0
    local turboAV = turboRPM * constants.rpmToAV
    
    self.motorAV = turboAV / self.motorGearRatio
    
    local throttle = self.engine.requestedThrottle
    local desiredTorqueCoef = maprange(throttle, 0, 0.01, -1, 0)
    local maxTorque = self:getMaxMotorTorque(desiredTorqueCoef < 0)
    local desiredTorque = desiredTorqueCoef * maxTorque
    local targetTorque = self.torqueSmoother:get(desiredTorque, dt)
    local effectiveTorque = 0
    
    if targetTorque < 0 then
        -- Turbo regen
        effectiveTorque = -sign(self.motorAV) * min(abs(targetTorque), self:getMaxMotorRegenTorque())
    else
        effectiveTorque = min(self:getMaxMotorTorque(), targetTorque)
    end
    
    self.torque = effectiveTorque
    self.power = mathutil.torqueToKw(self.torque, self.motorAV)
    
    -- Apply motor torque to turbo
    local turboTorque = self.torque / self.motorGearRatio
    
    self.turbo.setExternalTorque(turboTorque)
    
    debugutil:putValue("Turbo AV (RPM)", fnum(turboAV) .. " (" .. fnum(turboRPM) .. ")")
    debugutil:putValue("Motor AV (RPM)", fnum(self.motorAV) .. " (" .. fnum(self.motorAV * constants.avToRPM) .. ")")
    debugutil:putValue("Desired Torque", desiredTorque, 2)
    debugutil:putValue("Torque", self.torque, 2)
    debugutil:putValue("Power", self.power, 2)
    debugutil:putValue("Max Power", self.maxPower, 2)
end

function prespooler:setMaxPower(maxPower)
    self.maxPower = maxPower
end

function prespooler:getMaxMotorTorque(ignoreMaxPower)
    local av = max(abs(self.motorAV), 1e-6)
    local maxPower = ignoreMaxPower and self.powerRating or min(self.powerRating, self.maxPower)
    local maxTorque = mathutil.kwToTorque(maxPower, av)

    return min(maxTorque, self.torqueRating)
end

function prespooler:getMaxMotorRegenTorque()
    return maprange(abs(self.motorAV), 0, self.maxRegenMinAV, 0, self:getMaxMotorTorque())
end

function prespooler:reset()
    self.torqueSmoother:reset()
    self.targetTurboTorque = 0
    self.torque = 0
    self.power = 0
end

function M.new(jbeamData)
    local device = {
        jbeamData = jbeamData,
        isExisting = true,
        
        maxPower = 0,
        torque = 0,
        power = 0,
        motorAV = 0,
        
        motorGearRatio = jbeamData.motorGearRatio or 8.0,
        powerRating = jbeamData.powerRating or 25,
        torqueRating = jbeamData.torqueRating or 85,
        maxRegenMinAV = jbeamData.maxRegenMinAV or 30,
        maxEffectiveAV = jbeamData.maxEffectiveAV or 8000,
        
        torqueSmoother = newTemporalSmoothingNonLinear(
            jbeamData.torqueSmoothingIn or 12,
            jbeamData.torqueSmoothingOut or 12
        ),
    }
    
    setmetatable(device, prespooler)
    
    -- This function basically disables the device in the case of an error, but still returns it so the Lua engine doesn't crash
    local function fail(str)
        debugutil:log("E", "Error initializing turbo regen: " .. str)
        device.updateGFX = nop
        return device
    end
    
    device.engine = powertrain.getDevice("mainEngine")
    
    if not device.engine then
        return fail("Engine not found")
    end
    
    device.turbo = device.engine.turbocharger
    
    if not device.turbo then
        return fail("Turbo not found")
    end
    
    if type(device.turbo.setExternalTorque) ~= "function" then
        return fail("Turbo does not provide setExternalTorque function")
    end
    
    local turboJbeamName = device.engine.jbeamData.turbocharger
    
    if not turboJbeamName or not v.data[turboJbeamName] then
        return fail("Could not find turbocharger jBeam")
    end
    
    device.turboJbeam = v.data[turboJbeamName]
    device.turboElectricsRPMName = device.turboJbeam.electricsRPMName or "turboRPM"
    
    return device
end

return M